#include <iostream>
#include <chrono>
#include <functional>
#include <memory>
#include <vector>
#include <string>
#include <cstring>
#include <sstream>

#include <termios.h>
#include <fcntl.h>
#include <unistd.h>

#include <rclcpp/rclcpp.hpp>
#include <smc_3000_msgs/msg/drpva_a.hpp>

#define CRC32_POLYNOMIAL 0xEDB88320L

class Smc3000Node : public rclcpp::Node
{ 
public:
    Smc3000Node() : Node("smc3000") 
    {
        this->declare_parameter<std::string>("device", "/dev/ttyAMA1");
        this->declare_parameter<int>("baudrate", 115200);

        this->get_parameter("device", this->device);
        this->get_parameter("baudrate", this->baudrate);

        RCLCPP_INFO(this->get_logger(), "Device: %s, baudrate: %d", this->device.c_str(), this->baudrate);

        this->uart_fd = open(device.c_str(), O_RDWR | O_NOCTTY | O_NDELAY);

        if (this->uart_fd < 0) {
            RCLCPP_INFO(this->get_logger(), "Open %s error", this->device.c_str());
            return;
        }

        drpva_publisher = this->create_publisher<smc_3000_msgs::msg::DrpvaA>("smc3000/drpva", 10);

        tcgetattr(uart_fd, &this->options);

        switch (this->baudrate)
        {
        case 9600:
            this->options.c_cflag = B9600 | CS8 | CLOCAL | CREAD;
            break;

        case 38400:
            this->options.c_cflag = B38400 | CS8 | CLOCAL | CREAD;
            break;
        
        case 115200:
            this->options.c_cflag = B115200 | CS8 | CLOCAL | CREAD;
            break;
        
        case 230400:
            this->options.c_cflag = B230400 | CS8 | CLOCAL | CREAD;
            break;

        case 460800:
            this->options.c_cflag = B460800 | CS8 | CLOCAL | CREAD;
            break;

        case 921600:
            this->options.c_cflag = B921600 | CS8 | CLOCAL | CREAD;
            break;
        }

        this->options.c_iflag = IGNPAR;
        this->options.c_oflag = 0;
        this->options.c_lflag = 0;
        tcflush(this->uart_fd, TCIFLUSH);
        tcsetattr(this->uart_fd, TCSANOW, &this->options);

        RCLCPP_INFO(this->get_logger(), "Port READY");
        RCLCPP_INFO(this->get_logger(), "Timer START");

        data = "";

        uart_timer = this->create_wall_timer(
                     std::chrono::milliseconds(1),
                     std::bind(&Smc3000Node::get_smc3000_data, this));
    }

    ~Smc3000Node()
    {
        close(uart_fd);
    }

private:
    std::string device;
    int baudrate;
    int uart_fd;
    struct termios options;

    std::string data;

    smc_3000_msgs::msg::DrpvaA drpva_msg;
    rclcpp::Publisher<smc_3000_msgs::msg::DrpvaA>::SharedPtr drpva_publisher;

    rclcpp::TimerBase::SharedPtr uart_timer;

    void get_smc3000_data();
    bool parse_smc3000_data();
    unsigned long CRC32Value(int i);
    unsigned long CalculateBlockCRC32( unsigned long ulCount, const char *ucBuffer );
};

int main(int argc, char **argv)
{
    rclcpp::init(argc, argv);
    auto node = std::make_shared<Smc3000Node>();

    rclcpp::spin(node);
    rclcpp::shutdown();

    return 0;
}

void Smc3000Node::get_smc3000_data()
{
    // DRPVAA default length > 400
    char buffer[1024];
    int n = read(this->uart_fd, buffer, sizeof(buffer) - 1);
    if (n > 0) {
        buffer[n] = '\0';
        RCLCPP_INFO(this->get_logger(), "Received: %s", buffer);

        data += buffer;
        if (parse_smc3000_data()) {
            drpva_publisher->publish(drpva_msg);
        }
    } 
}

bool Smc3000Node::parse_smc3000_data()
{
    std::vector<std::string> splittedData;
    std::size_t startPos = this->data.find("#");
    if (startPos == std::string::npos) {
        // std::cout << "No # >> NO DRPVA" << std::endl;
        return false;
    } 

    std::size_t headerEndPos = this->data.find(";", startPos);
    if (headerEndPos == std::string::npos) {
        // std::cout << "No ; >> NO DRPVA" << std::endl;
        return false;
    }

    std::size_t drpvaEndPos = this->data.find("*", startPos);
    if (drpvaEndPos == std::string::npos || drpvaEndPos + 10 >= this->data.size()) {
        // std::cout << "No * >> NO DRPVA" << std::endl;
        return false;
    }
    
    size_t datalen = drpvaEndPos - startPos - 1;
    unsigned long CRC = CalculateBlockCRC32(datalen, data.substr(startPos + 1, datalen).c_str());

    std::string checkSumStr = this->data.substr(drpvaEndPos + 1, 8);
    unsigned long checksumProvided;
    std::stringstream ss;
    ss << std::hex << checkSumStr;
    ss >> checksumProvided;

    if (CRC != checksumProvided) {
        return false;
    }

    this->data = this->data.substr(drpvaEndPos + 9);

    splittedData.push_back("");
    for (size_t i = startPos + 1; i < drpvaEndPos; i++) {
        if (this->data[i] == ',' || this->data[i] == ';') splittedData.push_back("");
        else splittedData.back() += this->data[i];
    }

    // std::cout << splittedData[0] << std::endl;
    if (splittedData[0] != "DRPVAA") return false;

    
    int i = 0;
    drpva_msg.header.sync = splittedData[i++];
    drpva_msg.header.message = splittedData[i++];
    drpva_msg.header.port = std::stoi(splittedData[i++]);
    // drpva_msg.header.sequence_number = std::stoi(splittedData[i++]);
    drpva_msg.header.idle_time = std::stof(splittedData[i++]);
    drpva_msg.header.time_status = splittedData[i++];
    drpva_msg.header.week = std::stoul(splittedData[i++]);
    drpva_msg.header.seconds = std::stof(splittedData[i++]);
    drpva_msg.header.receiver_status = std::stoul(splittedData[i++]);
    drpva_msg.header.bds_time = std::stoul(splittedData[i++]);
    drpva_msg.header.utc_time = std::stol(splittedData[i++]);

    drpva_msg.sol_status = splittedData[i++];
    drpva_msg.pos_type = splittedData[i++];
    drpva_msg.datum_id = splittedData[i++];
    drpva_msg.reserved[0] = std::stoi(splittedData[i++]);
    drpva_msg.reserved[1] = std::stoi(splittedData[i++]);
    drpva_msg.reserved[2] = std::stoi(splittedData[i++]);
    drpva_msg.reserved[3] = std::stoi(splittedData[i++]);
    drpva_msg.dr_age = std::stof(splittedData[i++]);
    drpva_msg.sol_age = std::stof(splittedData[i++]);
    drpva_msg.lat = std::stod(splittedData[i++]);
    drpva_msg.lon = std::stod(splittedData[i++]);
    drpva_msg.hgt = std::stod(splittedData[i++]);
    drpva_msg.undulation = std::stof(splittedData[i++]);
    drpva_msg.lat_sig = std::stof(splittedData[i++]);
    drpva_msg.lon_sig = std::stof(splittedData[i++]);
    drpva_msg.hgt_sig = std::stof(splittedData[i++]);
    drpva_msg.ve = std::stod(splittedData[i++]);
    drpva_msg.vn = std::stod(splittedData[i++]);
    drpva_msg.vu = std::stod(splittedData[i++]);
    drpva_msg.ve_sig = std::stof(splittedData[i++]);
    drpva_msg.vn_sig = std::stof(splittedData[i++]);
    drpva_msg.vu_sig = std::stof(splittedData[i++]);
    drpva_msg.heading = std::stod(splittedData[i++]);
    drpva_msg.pitch = std::stod(splittedData[i++]);
    drpva_msg.roll = std::stod(splittedData[i++]);
    drpva_msg.heading_sig = std::stof(splittedData[i++]);
    drpva_msg.pitch_sig = std::stof(splittedData[i++]);
    drpva_msg.roll_sig = std::stof(splittedData[i++]);
    drpva_msg.reserved1[0] = std::stol(splittedData[i++]);
    drpva_msg.reserved1[1] = std::stol(splittedData[i++]);
    drpva_msg.reserved1[2] = std::stol(splittedData[i++]);
    drpva_msg.reserved1[3] = std::stol(splittedData[i++]);
    drpva_msg.reserved2[0] = std::stof(splittedData[i++]);
    drpva_msg.reserved2[1] = std::stof(splittedData[i++]);
    drpva_msg.reserved2[2] = std::stof(splittedData[i++]);
    drpva_msg.reserved2[3] = std::stof(splittedData[i++]);
    drpva_msg.reserved3[0] = std::stod(splittedData[i++]);
    drpva_msg.reserved3[1] = std::stod(splittedData[i++]);
    drpva_msg.reserved3[2] = std::stod(splittedData[i++]);
    drpva_msg.reserved3[3] = std::stod(splittedData[i++]);

    return true;
}

unsigned long Smc3000Node::CRC32Value(int i) {
     int j;
     unsigned long ulCRC;
     ulCRC = i;
     for ( j = 8 ; j > 0; j-- ) {
          if ( ulCRC & 1 )
               ulCRC = ( ulCRC >> 1 ) ^ CRC32_POLYNOMIAL;
          else
               ulCRC >>= 1;
     }
     return ulCRC;
}

unsigned long Smc3000Node::CalculateBlockCRC32( unsigned long ulCount, const char *ucBuffer ) {
     unsigned long ulTemp1;
     unsigned long ulTemp2;
     unsigned long ulCRC = 0;

     while ( ulCount-- != 0 ) {

          ulTemp1 = ( ulCRC >> 8 ) & 0x00FFFFFFL;

          ulTemp2 = CRC32Value( ((int) ulCRC ^ *ucBuffer++ ) & 0xFF );

          ulCRC = ulTemp1 ^ ulTemp2;

     }

     return( ulCRC );
}